package cmdVM

import (
    "regexp"
    "errors"
    "strings"
    "encoding/json"
    "nexus-client/cmd"
    "nexus-common/vm"
    u "nexus-client/utils"
    g "nexus-client/globals"
    "github.com/google/uuid"
    "github.com/go-resty/resty/v2"
    "github.com/go-playground/validator/v10"
)

func printRegexUsage(c cmd.Command) {
    for _, desc := range c.GetDesc() {
        u.PrintlnErr(desc)
    }
    u.PrintlnErr("―――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――")
    u.PrintlnErr("USAGE: ",c.GetName()," [ID ...] [regex ...]")
    u.PrintlnErr("―――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――")
}

func printRegexUsageDetails() {
    u.PrintlnErr("The action only applies to VMs matching the specified IDs or regexes.")
    const usage string = `Any number of IDs or regexes can be specified.
The regex only matches the VM's name and is case-insensitive.
Regex examples:
"."   -> matches any VMs
"bla" -> matches any VMs containing "bla" in their name`
    u.PrintlnErr(usage)
}

// Returns a list of filtered VMs for a given route.
// Filters are based on patterns describing IDs or regexes.
// Patterns can either be:
// - any number of "VM IDs" (UUID)
// - any number of "VM names"
// - any combination of "VM IDs" and "VM names"
// - any regular expression (it applies to the VM's name only)
// Remark: the matching is case-insensitive
// Regular expression examples:
//   "."   -> matches everything
//   "bla" -> matches any VM name containing "bla"
func getFilteredVMs(route string, patterns []string) ([]vm.VMNetworkSerialized, error) {
    if len(patterns) < 1 {
        return nil, errors.New("At least one ID or regex must be specified")
    }

    client := g.GetInstance().Client
    host := g.GetInstance().Host

    var ids []string
    var regexes []string

    for _, pattern := range patterns {
        _, err := uuid.Parse(pattern)
        if err != nil {
            regexes = append(regexes, pattern)
        } else {
            ids = append(ids, pattern)
        }
    }

    resp, err := client.R().Get(host+route)
    if err != nil {
        return nil, err
    }

    vmsList := []vm.VMNetworkSerialized{}

    if resp.IsSuccess() {
        vms, err := deserializeVMs(resp)
        if err != nil {
            return nil, err
        }

        for _, vm := range vms {
            found := false
            for _, id := range ids {
                if id == vm.ID.String() {
                    vmsList = append(vmsList, vm)
                    found = true
                    break
                }
            }
            if found {
                continue
            }
            for _, regex := range regexes {
                match, err := regexp.MatchString(strings.ToLower(regex), strings.ToLower(vm.Name))
                if err != nil {
                    return nil, errors.New("Error matching \""+regex+"\": "+err.Error())
                } else {
                    if match {
                        vmsList = append(vmsList, vm)
                        break
                    }
                }
            }
        }
        return vmsList, nil
    } else {
        return nil, errors.New("Error: "+resp.Status()+": "+resp.String())
    }
}

// Returns a list of filtered VM credentials for a given route.
// Filters are based on patterns describing IDs or regexes.
// Patterns can either be:
// - any number of "VM IDs" (UUID)
// - any number of "VM names"
// - any combination of "VM IDs" and "VM names"
// - any regular expression (it applies to the VM's name only)
// Remark: the matching is case-insensitive
// Regular expression examples:
//   "."   -> matches everything
//   "bla" -> matches any VM name containing "bla"
func getFilteredVMCredentials(route string, patterns []string) ([]vm.VMAttachCredentialsSerialized, error) {
    if len(patterns) < 1 {
        return nil, errors.New("At least one ID or regex must be specified")
    }

    client := g.GetInstance().Client
    host := g.GetInstance().Host

    var ids []string
    var regexes []string

    for _, pattern := range patterns {
        _, err := uuid.Parse(pattern)
        if err != nil {
            regexes = append(regexes, pattern)
        } else {
            ids = append(ids, pattern)
        }
    }

    resp, err := client.R().Get(host+route)
    if err != nil {
        return nil, err
    }

    vmsList := []vm.VMAttachCredentialsSerialized{}

    if resp.IsSuccess() {
        vms, err := deserializeVMsCredentials(resp)
        if err != nil {
            return nil, err
        }

        for _, vm := range vms {
            found := false
            for _, id := range ids {
                if id == vm.ID.String() {
                    vmsList = append(vmsList, vm)
                    found = true
                    break
                }
            }
            if found {
                continue
            }
            for _, regex := range regexes {
                match, err := regexp.MatchString(strings.ToLower(regex), strings.ToLower(vm.Name))
                if err != nil {
                    return nil, errors.New("Error matching \""+regex+"\": "+err.Error())
                } else {
                    if match {
                        vmsList = append(vmsList, vm)
                        break
                    }
                }
            }
        }
        return vmsList, nil
    } else {
        return nil, errors.New("Error: "+resp.Status()+": "+resp.String())
    }
}

// Deserialize a list of VMs from an http response.
func deserializeVMs(resp *resty.Response) ([]vm.VMNetworkSerialized, error) {
    vms := []vm.VMNetworkSerialized{}
    if err := json.Unmarshal(resp.Body(), &vms); err != nil {
        return nil, err
    }
    return vms, nil
}

// Deserialize a single VM from an http response.
func deserializeVM(resp *resty.Response) (*vm.VMNetworkSerialized, error) {
    var vm *vm.VMNetworkSerialized = &vm.VMNetworkSerialized{}
    if err := json.Unmarshal(resp.Body(), vm); err != nil {
        return vm, err
    }
    if err := validator.New(validator.WithRequiredStructEnabled()).Struct(vm); err != nil {
        return nil, err
    }
    return vm, nil
}

// Deserialize a list of VM credentials from an http response.
func deserializeVMsCredentials(resp *resty.Response) ([]vm.VMAttachCredentialsSerialized, error) {
    vmsCreds := []vm.VMAttachCredentialsSerialized{}
    if err := json.Unmarshal(resp.Body(), &vmsCreds); err != nil {
        return nil, err
    }
    return vmsCreds, nil
}

// Deserialize a VM credentials from an http response.
func deserializeVMCredentials(resp *resty.Response) (*vm.VMAttachCredentialsSerialized, error) {
    var vmCreds vm.VMAttachCredentialsSerialized
    if err := json.Unmarshal(resp.Body(), &vmCreds); err != nil {
        return nil, err
    }
    if err := validator.New(validator.WithRequiredStructEnabled()).Struct(vmCreds); err != nil {
        return nil, err
    }
    return &vmCreds, nil
}

